import logging
import os
import pickle

from autofit import conf
from autofit import exc

logger = logging.getLogger(__name__)


class ResultsCollection(object):
    def __init__(self):
        """
        A collection of results from previous phases. Results can be obtained using an index or the name of the phase
        from whence they came.
        """
        self.__result_list = []
        self.__result_dict = {}

    @property
    def last(self):
        """
        The result of the last phase
        """
        if len(self.__result_list) > 0:
            return self.__result_list[-1]
        return None

    @property
    def first(self):
        """
        The result of the first phase
        """
        if len(self.__result_list) > 0:
            return self.__result_list[0]
        return None

    def add(self, phase_name, result):
        """
        Add the result of a phase.

        Parameters
        ----------
        phase_name: str
            The name of the phase
        result
            The result of that phase
        """
        if phase_name in self.__result_dict:
            raise exc.PipelineException(
                "Results from a phase called {} already exist in the pipeline".format(phase_name))
        self.__result_list.append(result)
        self.__result_dict[phase_name] = result

    def __getitem__(self, item):
        """
        Get the result of a previous phase by index

        Parameters
        ----------
        item: int
            The index of the result

        Returns
        -------
        result: Result
            The result of a previous phase
        """
        return self.__result_list[item]

    def __len__(self):
        return len(self.__result_dict)

    def from_phase(self, phase_name):
        """
        Returns the result of a previous phase by its name

        Parameters
        ----------
        phase_name: str
            The name of a previous phase

        Returns
        -------
        result: Result
            The result of that phase

        Raises
        ------
        exc.PipelineException
            If no phase with the expected result is found
        """
        try:
            return self.__result_dict[phase_name]
        except KeyError:
            raise exc.PipelineException("No previous phase named {} found in results ({})".format(phase_name, ", ".join(
                self.__result_dict.keys())))


class Pipeline(object):

    def __init__(self, pipeline_name, *phases):
        """
        A pipeline of phases to be run sequentially. Results are passed between phases. Phases must have unique names.

        Parameters
        ----------
        pipeline_name: str
            The phase_name of this pipeline
        """
        self.pipeline_name = pipeline_name
        self.phases = phases
        phase_names = [phase.phase_name for phase in phases]
        if len(set(phase_names)) < len(phase_names):
            raise exc.PipelineException(
                "Cannot create pipelines with duplicate phase names. ({})".format(", ".join(phase_names)))

    def __add__(self, other):
        """
        Compose two runners

        Parameters
        ----------
        other: Pipeline
            Another pipeline

        Returns
        -------
        composed_pipeline: Pipeline
            A pipeline that runs all the  phases from this pipeline and then all the phases from the other pipeline
        """
        return self.__class__("{} + {}".format(self.pipeline_name, other.pipeline_name), *(self.phases + other.phases))

    def save_metadata(self, phase, data_name):
        """
        Save metadata associated with the phase, such as the name of the pipeline, the name of the phase and the name
        of the data being fit
        """
        with open("{}/.metadata".format(make_path(phase)), "w+") as f:
            f.write("pipeline={}\nphase={}\ndata={}".format(self.pipeline_name, phase.phase_name,
                                                            data_name))

    def run_function(self, func, data_name=None):
        """
        Run the function for each phase in the pipeline.

        Parameters
        ----------
        data_name
        func
            A function that takes a phase and prior results, returning results for that phase

        Returns
        -------
        results: ResultsCollection
            A collection of results
        """
        results = ResultsCollection()
        for i, phase in enumerate(self.phases):
            logger.info("Running Phase {} (Number {})".format(phase.optimizer.phase_name, i))
            assert_optimizer_pickle_matches_for_phase(phase)
            save_optimizer_for_phase(phase)
            self.save_metadata(phase, data_name)
            results.add(phase.phase_name, func(phase, results))
        return results


def make_optimizer_pickle_path(phase) -> str:
    """
    Create the path at which the optimizer pickle should be saved
    """
    return "{}/.optimizer.pickle".format(make_path(phase))


def make_path(phase) -> str:
    """
    Create the path to the folder at which the metadata and optimizer pickle should be saved
    """
    return "{}/{}{}{}".format(conf.instance.output_path, phase.phase_path, phase.phase_name, phase.phase_tag)


def save_optimizer_for_phase(phase):
    """
    Save the optimizer associated with the phase as a pickle
    """
    with open(make_optimizer_pickle_path(phase), "w+b") as f:
        f.write(pickle.dumps(phase.optimizer))


def assert_optimizer_pickle_matches_for_phase(phase):
    """
    Assert that the previously saved optimizer is equal to the phase's optimizer if a saved optimizer is found.

    Parameters
    ----------
    phase
        The phase

    Raises
    -------
    exc.PipelineException
    """
    path = make_optimizer_pickle_path(phase)
    if os.path.exists(path):
        with open(path, "r+b") as f:
            loaded_optimizer = pickle.loads(f.read())
            if phase.optimizer != loaded_optimizer:
                raise exc.PipelineException(
                    f"Can't restart phase at path {path} because settings don't match. "
                    f"Did you change the optimizer settings or model?")



import os

import pytest

from autofit import exc
from autofit import mock
from autofit.mapper import prior as p
from autofit.optimize import non_linear
from autofit.tools import pipeline


@pytest.fixture(name="results")
def make_results_collection():
    results = pipeline.ResultsCollection()

    results.add("first phase", "one")
    results.add("second phase", "two")

    return results


class TestResultsCollection(object):
    def test_with_name(self, results):
        assert results.from_phase("first phase") == "one"
        assert results.from_phase("second phase") == "two"

    def test_with_index(self, results):
        assert results[0] == "one"
        assert results[1] == "two"
        assert results.first == "one"
        assert results.last == "two"
        assert len(results) == 2

    def test_missing_result(self, results):
        with pytest.raises(exc.PipelineException):
            results.from_phase("third phase")

    def test_duplicate_result(self, results):
        with pytest.raises(exc.PipelineException):
            results.add("second phase", "three")


class MockPhase(object):
    def __init__(self, phase_name, optimizer=None):
        self.phase_name = phase_name
        self.optimizer = optimizer
        self.phase_path = phase_name
        self.phase_tag = phase_name


class TestPipeline(object):
    def test_unique_phases(self):
        pipeline.Pipeline("name", MockPhase("one"), MockPhase("two"))
        with pytest.raises(exc.PipelineException):
            pipeline.Pipeline("name", MockPhase("one"), MockPhase("one"))

    def test_optimizer_assertion(self):
        optimizer = non_linear.NonLinearOptimizer("Phase Name")
        optimizer.variable.profile = mock.GeometryProfile
        phase = MockPhase("phase_name", optimizer)

        try:
            os.makedirs(pipeline.make_path(phase))
        except FileExistsError:
            pass

        pipeline.save_optimizer_for_phase(phase)
        pipeline.assert_optimizer_pickle_matches_for_phase(phase)

        optimizer.variable.profile.centre_0 = p.UniformPrior()

        with pytest.raises(exc.PipelineException):
            pipeline.assert_optimizer_pickle_matches_for_phase(phase)



import logging
import shutil
from os import path

import pytest

from autofit import mock
from autofit.tools import phase as p
from autofit.tools import phase_property
from autofit.optimize import grid_search as gs
from autofit.optimize import non_linear

logger = logging.getLogger(__name__)

try:
    output_dir = "{}/../../workspace/output".format(path.dirname(path.realpath(__file__)))
    logger.info("Removing {}".format(output_dir))
    shutil.rmtree(output_dir)
except FileNotFoundError:
    logging.info("Not found")


class Analysis(non_linear.Analysis):
    def fit(self, instance):
        return -(instance.profile.centre[0] ** 2 + instance.profile.centre[1] ** 2)

    def visualize(self, instance, **kwargs):
        pass

    def describe(self, instance):
        return "{}, {}".format(*instance.profile.centre)


class Phase(p.AbstractPhase):
    profile = phase_property.PhaseProperty("profile")
    constant_profile = phase_property.PhaseProperty("constant_profile")

    def __init__(self, phase_name, phase_tag, phase_folders, profile, constant_profile, optimizer_class=non_linear.MultiNest):
        super().__init__(phase_name=phase_name, phase_tag=phase_tag, phase_folders=phase_folders,
                         optimizer_class=optimizer_class)
        self.profile = profile
        self.constant_profile = constant_profile

    def make_result(self, result, analysis):
        return result


class TestCase(object):
    
    def test_integration(self):
        multinest = non_linear.MultiNest(phase_folders=['integration'], phase_name='test')

        multinest.variable.profile = mock.EllipticalProfile

        result = multinest.fit(Analysis())

        centre = result.constant.profile.centre

        assert 0 == pytest.approx(centre[0], abs=0.1)
        assert 0 == pytest.approx(centre[1], abs=0.1)

    def test_grid(self):

        grid_search = gs.GridSearch(phase_name="phase_grid_search", phase_tag='tag', phase_folders=['integration'])
        grid_search.variable.profile = mock.EllipticalProfile

        # noinspection PyUnresolvedReferences
        result = grid_search.fit(Analysis(),
                                 [grid_search.variable.profile.centre_0, grid_search.variable.profile.centre_1])

    def test_phase(self):
        
        phase = Phase( phase_name="test_phase", phase_tag='tag', phase_folders=['integration'],
                       profile=mock.EllipticalProfile, constant_profile=mock.EllipticalProfile())
        result = phase.run_analysis(Analysis())

        centre = result.constant.profile.centre

        assert 0 == pytest.approx(centre[0], abs=0.1)
        assert 0 == pytest.approx(centre[1], abs=0.1)

    def test_classic_grid_search_phase(self):
        # noinspection PyTypeChecker
        phase = Phase(phase_name="phase_classic_grid_search_phase", phase_tag='tag', phase_folders=['integration'],
                      profile=mock.EllipticalProfile, constant_profile=mock.EllipticalProfile(),
                      optimizer_class=non_linear.GridSearch)
        result = phase.run_analysis(Analysis())

        centre = result.constant.profile.centre

        assert 0 == pytest.approx(centre[0], abs=0.1)
        assert 0 == pytest.approx(centre[1], abs=0.1)

    def test_grid_search_phase(self):
        class GridSearchPhase(p.as_grid_search(Phase)):
            @property
            def grid_priors(self):
                return [self.variable.profile.centre_0, self.variable.profile.centre_1]

        constant_profile = mock.EllipticalProfile()

        result = GridSearchPhase(phase_name="grid_search_phase", phase_tag='tag', phase_folders=['integration'],
                                 number_of_steps=2, profile=mock.EllipticalProfile,
                                 constant_profile=constant_profile).run_analysis(Analysis())

        assert result.results[0].constant.constant_profile == constant_profile

        print(result.figure_of_merit_array)

        assert result.figure_of_merit_array[0, 0] > result.figure_of_merit_array[0, 1]
        assert result.figure_of_merit_array[0, 0] > result.figure_of_merit_array[1, 0]
        assert result.figure_of_merit_array[1, 0] > result.figure_of_merit_array[1, 1]
        assert result.figure_of_merit_array[0, 1] > result.figure_of_merit_array[1, 1]



if __name__ == "__main__":
    TestCase().test_integration()
    TestCase().test_grid()
    TestCase().test_phase()
    TestCase().test_classic_grid_search_phase()
    TestCase().test_grid_search_phase()
